/*
	Piotr Romaski, G1ISI
	2008-03-25

	Zad. 16.
	Wzorzec Wizytator - przykad uycia

	Opis:
	Wzorzec Wizytatora jest wzorcem czynnociowym, ktry oddziela struktur
	danych w klasie od algorytmw.
	- moemy dodawa now funkcjonalno do istniejcych klas, nie
		modyfikujc przy tym tej klasy,
	- gromadzenie podobnych operacji w jednej klasie zamiast zmieniania lub
		dziedziczenia po klasie w celu dodania funckcjonalnoci,
	- ew. wsppraca ze Wzorcem Kompozycji.

	Wszelkie podobiestwo do kodu prezentowanego na wykadzie jest przypadkowe.
*/

// Uywane biblioteki
#include <iostream>
#define _USE_MATH_DEFINES  // Uywamy M_PI
#include <cmath>           // Uywane w liczeniu powierzchni

// Deklaracja klas definiowanych pniej, a uywanych w Visitor
class Square;
class Circle;

/*
	Abstrakcyjna klasa bazowa dla klas dostarczajcych nowej funkcjonalnoci.
	Zawiera wirtualne metody visit() dla kadej obsugiwanej przez ni klasy.
*/
class Visitor {
public:
	Visitor() { }
	virtual ~Visitor() { }

	virtual void visit(Square &) = 0;
	virtual void visit(Circle &) = 0;
};

/*
	Abstrakcyjna klasa bazowa dla klas przechowujcych dane.
	Zawiera wirtualn metod accept(), ktra docelowo ma "zaprasza"
	wizytatora  wywoywa jego metod visit() dla siebie lub obiektw
	w niej agregowanych.
*/
class Figure {
public:
	Figure() {}
	virtual ~Figure() {}

	virtual void accept(Visitor &) = 0;
};

/*
	Klasa przechowujca dane oraz implementujca metod accept()
	przez "zaproszenie" do siebie wizytatora.
*/
class Square : public Figure {
public:
	Square(double x1, double y1, double x2, double y2) :
	  x1_(x1), y1_(y1), x2_(x2), y2_(y2) {}
	virtual ~Square() {}

	virtual void accept(Visitor &visitor) {
		// Zaprasza do siebie wizytatora
		visitor.visit(*this);
	}

	// Brak enkapsulacji dla przejrzystoci przykadu
	double x1_,y1_;  // Jeden z wierzchokw
	double x2_,y2_;  // Przeciwlegy wierzchoek
};

/*
	Kolejna klasa przechowujca dane oraz implementujca metod accept
	przez "zaproszenie" do siebie wizytatora.
*/
class Circle : public Figure {
public:
	Circle(double x, double y, double r) : x_(x), y_(y), r_(r) {}
	virtual ~Circle() {}

	virtual void accept(Visitor &visitor) {
		// Zaprasza do siebie wizytatora
		visitor.visit(*this);
	}

	// Brak enkapsulacji dla przejrzystoci przykadu
	double x_, y_; // Wsprzdne rodka koa
	double r_;    // Promie koa
};

/*
	Wizytator, ktra "odwiedza" dane obiekty i wypisuje ich dane.
	W ten sposb atwo mona przeglda obiekty zawierajce struktury danych.
*/
class PrintDataVisitor : public Visitor {
public:
	PrintDataVisitor() { }
	virtual ~PrintDataVisitor() { }

	// Wypisuje w jednej linii dane kwadratu
	// Wyjcie: Square: v1=(x1,y1), v2=(x2,y2)
	virtual void visit(Square &square) {
		std::cout << "Square: v1=(" << square.x1_ << "," << square.y1_
			<< "), v2=(" << square.x2_ << "," << square.y2_ << ")" << std::endl;
	}
	// Wypisuje w jednej linii dane koa
	// Wyjcie: Circle: O=(x,y), R=r
	virtual void visit(Circle &circle) {
		std::cout << "Circle: O=(" << circle.x_ << "," << circle.y_
			<< "), R=" << circle.r_ << std::endl;
	}
};

/*
	Wizytator, ktra "odwiedza" dane figury i wylicza ich pola powierzchni.
	Wyniki s dodawane do poprzednich wynikw i zapamitywane - wizytator stanowy.
*/
class CountAreasVisitor : public Visitor {
public:
	CountAreasVisitor() : squares_area_(0), circles_area_(0) { }
	virtual ~CountAreasVisitor() { }

	// Wyliczenie powierzchni S = 0.5*d^2 i zapamitanie wyniku
	virtual void visit(Square &square) {
		squares_area_ +=
			0.5 * ( pow(square.x2_ - square.x1_, 2) + pow(square.y2_ - square.y1_, 2) );
	}
	// Wyliczenie powierzchni S = PI*R^2 i zapamitanie wyniku
	virtual void visit(Circle &circle) {
		circles_area_ += M_PI * pow(circle.r_, 2);
	}

	// Wypisanie rezultatw
	void print() const {
		std::cout << "Total area of squares: " << squares_area_ << std::endl;
		std::cout << "Total area of circles: " << circles_area_ << std::endl;
		std::cout << "Total area of squares and circles: " <<
			squares_area_+circles_area_ << std::endl;
	}

protected:
	double squares_area_;  // Suma pl odwiedzonych kwadratw
	double circles_area_;  // Suma pl odwiedzonych k
};


int main()
{
	// Tworzenie testowych danych
	// Wszystkie obiekty s wskazywane przez wskaniki klasy bazowej - Figure
	Figure *square1 = new Square(0.0,0.0,3.0,0.0);
	Figure *square2 = new Square(0.0,0.0,4.0,0.0);
	Figure *circle1 = new Circle(1.0,1.0,2);
	Figure *circle2 = new Circle(1.0,1.0,3);

	/*
		Zasada dziaania:
		metoda accept z obiektu figury z przekazaniem wizytatora =>
		=> metoda visit z podanego wizytatora z przekazaniem
		referencji na dan figur
	*/
	// Wypisanie danych wszystkich figur
	// !!! Brak sprawdzania typu danego obiektu
	PrintDataVisitor pdv;
	square1->accept(pdv); // Wyjcie: Square: v1=(0,0), v2=(3,0)
	square2->accept(pdv); // Wyjcie: Square: v1=(0,0), v2=(4,0)
	circle1->accept(pdv); // Wyjcie: Circle: O=(1,1), R=2
	circle2->accept(pdv); // Wyjcie: Circle: O=(1,1), R=3

	// Wyliczenie pl powierzchni wszystkich figur
	// Wyniki s sumowane i przechowywane w wizytatorze - zachowanie stanu
	CountAreasVisitor cav;
	square1->accept(cav);
	square2->accept(cav);
	circle1->accept(cav);
	circle2->accept(cav);
	cav.print();
	/*
	 Wyjcie:
	 Total area of squares: 12.5
	 Total area of circles: 40.8407
	 Total area of squares and circles: 53.3407
	*/

	// Usuwanie danych testowych
	delete square1;	delete square2;
	delete circle1;	delete circle2;

	return 0;
}
